/* 
 * Copyright (C) International Computer Science Institute, 1995/1996.  COPYRIGHT
 * NOTICE: This code is provided "AS IS" WITHOUT ANY WARRANTY and is subject
 * to the terms of the SATHER LIBRARY GENERAL PUBLIC LICENSE contained in
 * the file "Doc/License" of the Sather distribution.  The license is also
 * available from ICSI, 1947 Center St., Suite 600, Berkeley CA 94704, USA.
 * ---> Please email bugs to "psather-bugs@icsi.berkeley.edu". 
 *
 * General pSather runtime stuff
 *
 * Version 1.0 (released for 1.0.9) by Claudio Fleiner (fleiner@icsi.berkeley.edu)
 */
#include <stddef.h>
#include "memory.h"
#include "debug.h"
#include "locks.h"
#include "trace.h"
#include "stat.h"
#include <am.h>

extern void p_init_locks(void);
extern void init_thread(void);
char *bin_prog_name=NULL;
long start_sec;
extern void deamon_thread();

#if defined(_REENTRANT) && defined(__GNUC__) && defined(__sparc__)
/* use spinlocks */
spinlock_t vlocks[VLOCKS];
#else
lock_t vlocks[VLOCKS];
#endif

#ifdef DEFAULT_CLUSTERS
int clusters=DEFAULT_CLUSTERS;
#else
int clusters=4;
#endif

/* used if backtracing in pSather in on */
CREATE_FF

/*
 * When the following function is called from within the debugger,
 * it will start a new thread that calls "debug_psather()" once
 * every second. This way one is sure to get a running thread 
 * while debugging pSather
 */
void debug_psather_request(vnn_t from)
{
	extern void get_thread();
	while(1) {
		sleep(1);
		get_thread();
	}
}

void get_thread() {}
void stop_thread() { lock_t l=LCK_CREATE;while(1) LCK_LOCK(l); }
void debug_psather_thread()
{
	int i;
	/*
	for(i=0;i<CLUSTERS;i++)
		thr_create_0(i,debug_psather_request);
	*/
}

static void end_psather_done(vnn_t from,long arg,long lck)
{
	LCK_LOCK((lock_t)lck);
	(*(long *)arg)--;
	LCK_UNLOCK((lock_t)lck);
}

static void end_psather(vnn_t from,long arg,long lck)
{
	am_request_2(from,end_psather_done,arg,lck);
	stat_print(bin_prog_name);
	psather_stop();
	TRACE("cluster died");
	exit(0);
}

void set_startsec(vnn_t from,long s)
{
	start_sec=s;
}

void correct_startsec()
{
	int i;

	for(i=1;i<CLUSTERS;i++)
		am_request_1(i,set_startsec,start_sec);
}

/* makes sure that the lock manager has the correct local 
   thread memory */
static void init_poll_manager(vnn_t from)
{
	extern void init_thread_memory();
	init_thread_memory();
	if(HERE==1) am_reply_0(0,init_poll_manager);
}


void psather_start(int argc,char *argv[])
{
	int i;
	volatile int j=1;
	extern void init_import_export();
	extern int gettimeofday(struct timeval *,void *);
	struct timeval now;
	extern void init_thread_count();
	extern void gdb_signal_handler(int);
	extern void remote_gdb(int);

	extern int sscanf(const char *,const char *,...);
	/*
	do {
		gettimeofday(&now,NULL);
	} while(now.tv_usec>10000);
	*/
	gettimeofday(&now,NULL);
	start_sec=now.tv_sec;

        signal(SIGQUIT,remote_gdb);
        signal(SIGILL,gdb_signal_handler);
        signal(SIGEMT,gdb_signal_handler);
        signal(SIGFPE,gdb_signal_handler);
        signal(SIGBUS,gdb_signal_handler);
        signal(SIGSEGV,gdb_signal_handler);
        signal(SIGSYS,gdb_signal_handler);

	init_psather_prio();

	bin_prog_name=argv[0];
#ifdef PSATHER_STAT
	init_stat();
#endif
#if defined(_REENTRANT) && defined(__GNUC__) && defined(__sparc__)
	/* use spinlocks */
	for(j=0;j<VLOCKS;j++) vlocks[j]=0; /* initialize vlocks */
#else
	for(j=0;j<VLOCKS;j++) vlocks[j]=LCK_CREATE; /* initialize vlocks */
#endif
#ifdef PSATHER1
	clusters=1;
#else
	if(getenv("CLUSTERS")!=NULL) {
		sscanf(getenv("CLUSTERS"),"%d",&clusters);
	}
#endif
	init_thread_count();
	init_import_export(); /* initialize import/export stuff */
	init_parloop_locks();
	am_enable(CLUSTERS,argc,argv);
	init_psather_end_detection();
	init_parloops();
	clusters=am_procs();
	init_thread();

	TRACE2("cluster started at %d.%06d",now.tv_sec,now.tv_usec);
	p_init_locks();
	/* initilaize registartion database */
	REG(NULL,NULL);UNREG;
	if(am_my_proc()!=0) {
		lock_t t;
		debug_psather_thread(); /* only needed for debugging */
		t=LCK_CREATE;
		deamon_thread();
		while(1) LCK_LOCK(t);
	}

	correct_startsec();
	for(i=1;i<CLUSTERS;i++)
		am_request_0(i,init_poll_manager);
}

int psather_stop(void)
{
	int i;
	long j=0;
	lock_t m;
	extern lock_t psather_dead;
	m=lck_create();
	if(am_my_proc()==0) {
		deamon_thread();
		LCK_LOCK(psather_dead);
		j=CLUSTERS-1;
		for(i=1;i<CLUSTERS;i++) {
			thr_create_2(i,end_psather,(long)&j,(long)m);
		}
		stat_print(bin_prog_name);
		TRACE("cluster died");
	}
	am_wait_for(j==0);
	am_disable();
	return 0;
}

int read_tag(FOB obj)
{
	short i;
	READTAG(i,obj);
	return i;
}
#undef abort
int psather_abort(void) { abort(); return 0; }

int p_fwrite(FOB p,int size,int elems,FILE *file)
{
	void *x;
	int i;
	if(!NEAR(p)) {
		TA_SEMAPHORE(a);
		MALLOC(x,size*elems);
		am_get(WHERE(p),SENDFOBHOME(p),x,size*elems,r_ta_sema_signal_mem,a);
		TA_SEMA_WAIT(a);
	} else x=p;
	i=fwrite(x,size,elems,file);
	if(!NEAR(p)) FREE(x);
	return i;
}
long p_sysid(FOB r)
{
	long x,y;

	SAFE_POS
#ifdef PSATHER_CHK
		if(F_TAG(r)<0) RFATAL("Called SYSID::id on value type");
#endif
		x=WHERE(r);
#ifdef DETERMINISTIC
		F_VA_RATTR_NA(y,OB,r,header.id);
#else
		y=((long)POINTER(r))>>3;
#endif
	RESTORE_POS
	return (x<<CLUSTER_SHIFT)+y;
}

	
static void request_sysdestroy(vnn_t from,FOB x)
{
#ifdef SYSDESTROY /* this cannot be tested without the Sather Runtime files */
	extern rt_fatal2();
	SYSDESTROY(x);
#endif
}
void p_sysdestroy(FOB x)
{
	am_request_1(WHERE(x),(handler_1_t)request_sysdestroy,(long)SENDFOBHOME(x));
}
int pmin(int a, int b)
{
	return a<b?a:b;
}

int p_rfatalv(char *s)
{
	if(THR_LOCAL==NULL) fprintf(stderr,"%s",s);
        else  {
		fprintf(stderr,"%s:%ld:%s\n",((LOCAL_MEM)THR_LOCAL)->file,(long)((LOCAL_MEM)THR_LOCAL)->line,s);
#ifdef PRINT_PO
		{
			extern void PT(),PF();
			extern int print_gdb;
			int pg;
			pg=print_gdb;
			if(getenv("START_GDB")==NULL) print_gdb=0;
			printf("------------------------------------\ncurrent function frame:\n");
			PF(((LOCAL_MEM)THR_LOCAL)->pFF);
			printf("------------------------------------\nbacktrace:\n");
			PT(((LOCAL_MEM)THR_LOCAL)->pFF);
			print_gdb=pg;
			printf("------------------------------------\n");
			fflush(stdout);
		}
#endif
	}
	start_gdb();
	PSATHER_ABORT;
	return 0;
}

int p_rfatalvfl(char *s,char *f,int l)
{
	DEBUG0(s);
	fprintf(stderr,"%s:%d:%s\n",f,l,s);
	fflush(stderr);
	start_gdb();
	PSATHER_ABORT;
	return 0;
}

void break_psather() { SYS_DEFER; } 
